package ru.dbotthepony.mc.otm.container

import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.ints.IntArraySet
import it.unimi.dsi.fastutil.ints.IntCollection
import it.unimi.dsi.fastutil.ints.IntList
import it.unimi.dsi.fastutil.ints.IntOpenHashSet
import it.unimi.dsi.fastutil.ints.IntSet
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenCustomHashMap
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import net.minecraft.world.Container
import net.minecraft.world.inventory.CraftingContainer
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.enchantment.EnchantmentEffectComponents
import net.minecraft.world.item.enchantment.EnchantmentHelper
import net.neoforged.neoforge.fluids.capability.IFluidHandler
import ru.dbotthepony.kommons.collect.addAll
import ru.dbotthepony.kommons.collect.filter
import ru.dbotthepony.kommons.collect.toList
import ru.dbotthepony.kommons.util.map
import ru.dbotthepony.mc.otm.container.util.ItemStackHashStrategy
import ru.dbotthepony.mc.otm.container.util.containerSlot
import ru.dbotthepony.mc.otm.container.util.slotIterator
import ru.dbotthepony.mc.otm.util.collect.IntRange2Set
import ru.dbotthepony.mc.otm.util.isNotEmpty
import ru.dbotthepony.mc.otm.util.ItemStackSorter
import kotlin.math.roundToInt

@Suppress("nothing_to_inline")
inline operator fun Container.set(index: Int, value: ItemStack) = setItem(index, value)

@Suppress("nothing_to_inline")
inline operator fun Container.get(index: Int): ItemStack = getItem(index)

@Suppress("nothing_to_inline")
inline operator fun IFluidHandler.get(index: Int) = getFluidInTank(index)

val Container.slotRange: IntRange2Set get() {
	return IntRange2Set.openEnded(0, containerSize)
}

fun Container.addItem(stack: ItemStack, simulate: Boolean, slots: IntSet = slotRange): ItemStack {
	return IEnhancedContainer.wrap(this).addItem(stack, simulate, slots)
}

fun Container.vanishCursedItems() {
	for (slot in slotIterator()) {
		if (EnchantmentHelper.has(slot.item, EnchantmentEffectComponents.PREVENT_EQUIPMENT_DROP)) {
			slot.remove()
		}
	}
}

fun Container.balance(slots: IntCollection, checkForEmpty: Boolean = true) {
	if (slots.isEmpty() || checkForEmpty && !slots.any { getItem(it).isNotEmpty }) return

	val empty = IntArrayList()
	val itemTypes = Object2ObjectOpenCustomHashMap<ItemStack, IntAVLTreeSet>(ItemStackHashStrategy)

	for (i in slots.intIterator()) {
		val item = getItem(i)

		if (item.isEmpty) {
			empty.add(i)
		} else {
			itemTypes.computeIfAbsent(item, Object2ObjectFunction { IntAVLTreeSet() }).add(i)
		}
	}

	if (itemTypes.isEmpty())
		return

	// только один вид предмета, просто балансируем его во все слоты
	if (itemTypes.size == 1) {
		val (item, list) = itemTypes.entries.first()
		var count = 0
		for (i in list.intIterator()) count += getItem(i).count
		var ghostChanges = false

		// всего предметов меньше, чем слотов
		if (count < slots.size) {
			for (slot in list.intIterator()) {
				getItem(slot).count = 1
				ghostChanges = true
			}

			count -= list.size

			while (count > 0 && empty.isNotEmpty()) {
				setItem(empty.removeInt(0), item.copyWithCount(1))
				count--
			}

			if (count > 0) {
				getItem(list.firstInt()).count += count
				ghostChanges = true
			}
		} else {
			// всего предметов больше, чем слотов
			val perSlot = count / slots.size
			var leftover = count - perSlot * slots.size

			for (i in slots.intIterator()) {
				val target = perSlot + leftover.coerceAtMost(1)

				if (getItem(i).count != target) {
					setItem(i, item.copyWithCount(target))
				}

				if (leftover > 0) leftover--
			}
		}

		if (ghostChanges) setChanged()
		return
	}

	// а вот тут уже проблемы
	// для упрощения задачи, выполним рекурсивное разбитие задачи на более простые,
	// где балансировка будет происходить только между пустыми слотами и предметами одного типа

	// если у нас нет пустых слотов, просто балансируем между заполненными слотами
	if (empty.isEmpty) {
		for (set in itemTypes.values) {
			balance(set)
		}

		return
	} else if (empty.size == 1) {
		// только один пустой слот, отдадим самому "жирному" предмету
		val type = itemTypes.entries.stream().max { a, b -> b.value.intStream().sum().compareTo(a.value.intStream().sum()) }.orElseThrow()
		type.value.add(empty.getInt(0))
		balance(type.value)

		for ((a, set) in itemTypes) {
			if (a !== type.key) {
				balance(set)
			}
		}

		return
	}

	// определяем общее количество предметов
	val totalCount = itemTypes.values.stream().mapToInt { it.stream().mapToInt { getItem(it).count }.sum() }.sum().toDouble()
	val totalEmpty = empty.size

	// определяем доли предметов по их количеству к общему количеству,
	// что позволит нам выделить претендентов на пустые слоты
	for (list in itemTypes.values) {
		if (empty.isEmpty) break // ошибка округления

		val perc = list.stream().mapToInt { getItem(it).count }.sum() / totalCount

		for (i in 0 until (perc * totalEmpty).roundToInt()) {
			list.add(empty.removeInt(0))
			if (empty.isEmpty) break // ошибка округления
		}
	}

	if (empty.isNotEmpty()) {
		// ошибка округления
		itemTypes.values.stream().max { a, b -> b.intStream().sum().compareTo(a.intStream().sum()) }.orElseThrow().add(empty.removeInt(0))
	}

	for (list in itemTypes.values) {
		balance(list)
	}
}

fun Container.balance(slots: Iterator<Int>) {
	balance(IntArraySet().also { it.addAll(slots) })
}

fun Container.balance(slots: Iterable<Int>) {
	balance(IntArraySet().also { it.addAll(slots) })
}

fun Container.balance(slots: IntRange) {
	balance(IntArraySet().also { it.addAll(slots) })
}

fun Container.balance(startSlot: Int, endSlot: Int) {
	require(startSlot <= endSlot) { "Invalid slot range: $startSlot .. $endSlot" }
	var any = false

	for (i in startSlot .. endSlot) {
		if (getItem(i).isNotEmpty) {
			any = true
			break
		}
	}

	if (!any) return
	balance(IntArraySet(endSlot - startSlot + 1).also { for (i in startSlot .. endSlot) it.add(i) }, false)
}

fun Container.balance() {
	if (isEmpty)
		return

	balance(IntArraySet(containerSize).also { for (i in 0 until containerSize) it.add(i) }, false)
}

operator fun CraftingContainer.get(column: Int, row: Int): ItemStack {
	return getItem(column + row * width)
}

operator fun CraftingContainer.get(column: Int, row: Int, flop: Boolean): ItemStack {
	return if (flop)
		get(width - column - 1, row)
	else
		get(column, row)
}

fun Container.sort(comparator: Comparator<ItemStack> = ItemStackSorter.DEFAULT) {
	sortWithIndices(computeSortedIndices(comparator))
}

fun Container.sortWithIndices(sortedSlots: IntCollection) {
	if (sortedSlots.isEmpty() || isEmpty)
		return

	val seen = IntOpenHashSet()
	val valid = ArrayList<IContainerSlot>()

	val iterator = sortedSlots.intIterator()

	while (iterator.hasNext()) {
		val value = iterator.nextInt()

		if (value in 0 until containerSize && seen.add(value)) {
			val slot = containerSlot(value)

			val condition: Boolean

			if (slot is IFilteredContainerSlot) {
				condition = slot.isNotEmpty &&
					!slot.filter.denyAll &&
					slot.item.count <= slot.maxStackSize(slot.item) &&
					(slot.filter.allowAll || !slot.filter.test(slot.item) || slot.maxStackSize(slot.item) > 1)
			} else {
				condition = slot.isNotEmpty && slot.item.count <= slot.maxStackSize(slot.item)
			}

			if (condition) {
				valid.add(slot)
			}
		}
	}

	if (valid.isEmpty())
		return

	val items = valid.map { it.item.copy() }
	valid.forEach { it.remove() }
	items.forEach { addItem(it, false) }
}

fun Container.computeSortedIndices(comparator: Comparator<ItemStack> = ItemStackSorter.DEFAULT): IntList {
	if (isEmpty)
		return IntList.of()

	val slots = slotIterator()
		.withIndex()
		.filter { (_, it) -> (it !is IFilteredContainerSlot || !it.filter.denyAll) && it.maxStackSize(it.item) >= it.item.count }
		.toList()

	if (slots.isEmpty())
		return IntList.of()

	val items = Object2ObjectOpenCustomHashMap<ItemStack, Pair<ItemStack, IntList>>(ItemStackHashStrategy)

	slots.forEach { (index, it) ->
		val get = items[it.item]

		if (get == null) {
			items[it.item] = it.item.copy() to IntArrayList().also { s -> s.add(index) }
		} else {
			get.first.count += it.item.count
			get.second.add(index)
		}
	}

	val sortedItems = ObjectArrayList(items.values)
	sortedItems.sortWith(comparator.map { it.first })

	val result = IntArrayList()
	sortedItems.forEach { result.addAll(it.second) }
	return result
}
