package ru.dbotthepony.mc.otm.container.slotted

import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.capabilities.Capabilities
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.fluid.stream
import ru.dbotthepony.mc.otm.container.IContainerSlot
import ru.dbotthepony.mc.otm.util.isNotEmpty
import ru.dbotthepony.mc.otm.util.math.Decimal

enum class AutomationFilters : AutomationFilter<IContainerSlot> {
	ONLY_OUT {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return true
		}
	},

	ONLY_IN {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return true
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return false
		}
	},

	ALLOW {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return true
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return true
		}
	},

	DENY {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return false
		}
	},

	FLUID_CONTAINERS {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getCapability(Capabilities.FluidHandler.ITEM)?.let { it.tanks > 0 } ?: false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return true
		}
	},

	DRAINABLE_FLUID_CONTAINERS {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getCapability(Capabilities.FluidHandler.ITEM)?.let { it.stream().anyMatch { it.isNotEmpty } } ?: false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return !canAutomationPlaceItem(self, self.item)
		}
	},

	DISCHARGABLE {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getCapability(Capabilities.EnergyStorage.ITEM)?.let { it.canExtract() && it.extractEnergy(Int.MAX_VALUE, true) > 0 } ?: false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return self.item.getCapability(Capabilities.EnergyStorage.ITEM)?.let { !it.canExtract() || it.extractEnergy(Int.MAX_VALUE, true) <= 0 } ?: true
		}
	},

	CHARGEABLE {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getCapability(Capabilities.EnergyStorage.ITEM)?.let { it.canReceive() && it.receiveEnergy(Int.MAX_VALUE, true) > 0 } ?: false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return self.item.getCapability(Capabilities.EnergyStorage.ITEM)?.let { !it.canReceive() || it.receiveEnergy(Int.MAX_VALUE, true) <= 0 } ?: true
		}
	},

	CHEMICAL_FUEL {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getBurnTime(null) > 0
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return self.item.getBurnTime(null) <= 0
		}
	},

	IS_PATTERN {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getCapability(MatteryCapability.PATTERN_ITEM) != null
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return true
		}
	},

	MATTER_PROVIDERS {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { it.matterFlow.output && it.extractMatterChecked(Decimal.POSITIVE_INFINITY, true) > Decimal.ZERO }
				?: false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return self.item.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { !it.matterFlow.output || it.extractMatterChecked(Decimal.POSITIVE_INFINITY, true) <= Decimal.ZERO }
				?: true
		}
	},

	MATTER_CONSUMERS {
		override fun canAutomationPlaceItem(self: IContainerSlot, itemStack: ItemStack): Boolean {
			return itemStack.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { it.matterFlow.input && it.receiveMatterChecked(Decimal.POSITIVE_INFINITY, true) > Decimal.ZERO }
				?: false
		}

		override fun canAutomationTakeItem(self: IContainerSlot, desired: Int): Boolean {
			return self.item.getCapability(MatteryCapability.MATTER_ITEM)
				?.let { !it.matterFlow.input || it.receiveMatterChecked(Decimal.POSITIVE_INFINITY, true) <= Decimal.ZERO }
				?: true
		}
	};

	val simpleProvider: SlottedContainer.SlotProvider<ContainerSlot> by lazy(LazyThreadSafetyMode.PUBLICATION) {
		ContainerSlot.Simple(filter = this)
	}

	val filteredProvider: SlottedContainer.SlotProvider<FilteredContainerSlot> by lazy(LazyThreadSafetyMode.PUBLICATION) {
		FilteredContainerSlot.Simple(filter = this)
	}

	val unlimitedSimpleProvider: SlottedContainer.SlotProvider<ContainerSlot> by lazy(LazyThreadSafetyMode.PUBLICATION) {
		ContainerSlot.Simple(filter = this, maxStackSize = Int.MAX_VALUE)
	}

	val unlimitedFilteredProvider: SlottedContainer.SlotProvider<FilteredContainerSlot> by lazy(LazyThreadSafetyMode.PUBLICATION) {
		FilteredContainerSlot.Simple(filter = this, maxStackSize = Int.MAX_VALUE)
	}

	val limitedSimpleProvider: SlottedContainer.SlotProvider<ContainerSlot> by lazy(LazyThreadSafetyMode.PUBLICATION) {
		ContainerSlot.Simple(filter = this, maxStackSize = 1)
	}

	val limitedFilteredProvider: SlottedContainer.SlotProvider<FilteredContainerSlot> by lazy(LazyThreadSafetyMode.PUBLICATION) {
		FilteredContainerSlot.Simple(filter = this, maxStackSize = 1)
	}
}
